package com.goldencode.p2j.testengine;

import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

import org.junit.jupiter.api.*;
import org.junit.platform.commons.support.*;
import org.junit.platform.commons.util.*;
import org.junit.platform.engine.*;
import org.junit.platform.engine.support.descriptor.*;
import org.junit.platform.engine.support.hierarchical.*;

/**
 * Class test descriptor.
 */
public class FWDClassTestDescriptor
extends AbstractTestDescriptor
implements Node<FWDEngineExecutionContext>
{

   /**
    * The Java class, referenced by child descriptors.
    */
   final Class<?> clazz;

   /**
    * The methods marked as {@link AfterAll}.
    */
   private final List<Method> afterAll;

   /**
    * The methods marked as {@link BeforeAll}.
    */
   private final List<Method> beforeAll;

   /**
    * Test a Java class candidate for existent tests methods. If found, create a new class
    * test descriptor.
    * 
    * @param clazz
    *        the class under test 
    * @param parentDescriptor
    *        the parent descriptor
    *        
    * @return  the class test descriptor or {@code null} if this class has no test methods.
    */
   public static FWDClassTestDescriptor testAndCreate(final Class<?> clazz,
            final TestDescriptor parentDescriptor)
   {
      // FIXME: find out if the test for abstract classes is really necessary.
      // For example, Eclipse does not run abstract classes as JUnit test
      if (Modifier.isAbstract(clazz.getModifiers()))
      {
         return null;
      }

      final List<Method> test = collectAnnotatedMethods(clazz, Test.class);
      if (test.isEmpty())
      {
         return null;
      }

      return new FWDClassTestDescriptor(clazz, parentDescriptor, test,
               collectAnnotatedMethods(clazz, BeforeEach.class),
               collectAnnotatedMethods(clazz, AfterEach.class),
               collectAnnotatedMethods(clazz, BeforeAll.class),
               collectAnnotatedMethods(clazz, AfterAll.class));
   }

   /**
    * Collect all methods annotated with the given class, but not annotated as disabled.
    * 
    * @param clazz
    *        the class to scan for methods
    * @param annoClass
    *        the annotation class to match
    * @return
    *        the list of methods found; never {@code null}
    */
   private static List<Method> collectAnnotatedMethods(final Class<?> clazz,
            final Class<? extends Annotation> annoClass)
   {
      return ReflectionUtils.findMethods(clazz, m -> AnnotationUtils.isAnnotated(m, annoClass)
               && !AnnotationUtils.isAnnotated(m, Disabled.class));
   }

   /**
    * Private constructor.
    * 
    * @param clazz
    *        the Java class
    * @param parent
    *        the parent descriptor
    * @param test
    *        class methods marked as tests  
    * @param beforeEach
    *        class methods marked to execute before each test  
    * @param afterEach
    *        class methods marked to execute after each test  
    * @param beforeAll
    *        class methods marked to execute before all tests  
    * @param afterAll
    *        class methods marked to execute aftger all tests  
    */
   private FWDClassTestDescriptor(final Class<?> clazz, final TestDescriptor parent,
            List<Method> test, List<Method> beforeEach, List<Method> afterEach,
            List<Method> beforeAll, List<Method> afterAll)
   {
      super(parent.getUniqueId().append("class", clazz.getName()), clazz.getSimpleName(),
               ClassSource.from(clazz));

      setParent(parent);

      this.clazz = clazz;
      this.beforeAll = beforeAll;
      this.afterAll = afterAll;
      test.stream().map(m -> new FWDMethodTestDescriptor(this, clazz, m, beforeEach, afterEach))
               .forEach(this::addChild);
   }

   /**
    * Execute the <em>after</em> behavior of this node.
    *
    * <p>This method will be called once <em>after</em> {@linkplain #execute
    * execution} of this node.
    *
    * <p>The default implementation does nothing.
    *
    * @param context the context to execute in
    * @see #before
    * @see #execute
    */
   @Override
   public void after(final FWDEngineExecutionContext context) throws Exception
   {
      for (Method m : afterAll)
      {
         // System.err.println("executing after class method " + m);
         ReflectionSupport.invokeMethod(m, null);
      }
   }

   /**
    * Execute the <em>before</em> behavior of this node.
    *
    * <p>This method will be called once <em>before</em> {@linkplain #execute
    * execution} of this node.
    *
    * <p>The default implementation returns the supplied {@code context} unmodified.
    *
    * @param context the context to execute in
    * @return the new context to be used for children of this node; never
    * {@code null}
    */
   @Override
   public FWDEngineExecutionContext before(FWDEngineExecutionContext context) throws Exception
   {
      for (Method m : beforeAll)
      {
         // System.err.println("executing before class method " + m);
         ReflectionSupport.invokeMethod(m, null);
      }

      return context;
   }

   /**
    * Execute the <em>behavior</em> of this node.
    *
    * <p>Containers typically do not implement this method since the
    * {@link HierarchicalTestEngine} handles execution of their children.
    *
    * <p>The supplied {@code dynamicTestExecutor} may be used to submit
    * additional dynamic tests for immediate execution.
    *
    * <p>The default implementation returns the supplied {@code context} unmodified.
    *
    * @param context the context to execute in
    * @param dynamicTestExecutor the executor to submit dynamic tests to
    * @return the new context to be used for children of this node and for the
    * <em>after</em> behavior of the parent of this node, if any
    * @see #before
    * @see #after
    */
   @Override
   public FWDEngineExecutionContext execute(FWDEngineExecutionContext context,
            DynamicTestExecutor dynamicTestExecutor) throws Exception
   {
      return context;
   }

   /**
    * Determine the {@link org.junit.platform.engine.TestDescriptor.Type} of this descriptor.
    *
    * @return the descriptor type; never {@code null}.
    * @see #isContainer()
    * @see #isTest()
    */
   @Override
   public Type getType()
   {
      return Type.CONTAINER;
   }

}
