package com.goldencode.p2j.testengine;

import java.lang.reflect.*;

import org.junit.platform.commons.support.*;
import org.junit.platform.engine.support.descriptor.*;
import org.junit.platform.engine.support.hierarchical.*;

/**
 * Method test descriptor.
 */
public class FWDMethodTestDescriptor
extends AbstractTestDescriptor
implements Node<FWDEngineExecutionContext>
{
   /**
    * The Java class to create instances of.
    * 
    */
   final Class<?> clazz;

   /**
    * Methods to execute after the main method, if any 
    */
   private final Iterable<Method> after;

   /**
    * Methods to execute before the main method, if any 
    */
   private final Iterable<Method> before;

   /**
    * Main method to execute
    */
   private final Method method;

   /**
    * The test class instance. Is created before each test life cycle begins,
    * and dropped after test life cycle ends.
    */
   private Object testInstance;

   /**
    * Constructor.
    * @param parent 
    *        the parent descriptor
    * @param clazz
    *        the Java class to create instances
    * @param execute
    *        the method to execute
    * @param before
    *        methods to execute before the execute method
    * @param after
    *        methods to execute after the execute method
    */
   FWDMethodTestDescriptor(final FWDClassTestDescriptor parent, Class<?> clazz,
            final Method execute, final Iterable<Method> before, final Iterable<Method> after)
   {
      super(parent.getUniqueId().append("method", execute.getName()), execute.getName(),
               new FWDMethodTestSource());

      this.clazz = clazz;
      this.method = execute;
      this.before = before;
      this.after = after;

      setParent(parent);
   }

   /**
    * Execute the <em>after</em> behavior of this node.
    *
    * <p>This method will be called once <em>after</em> {@linkplain #execute
    * execution} of this node.
    *
    * @param context
    *        the context to execute in
    * 
    * @see #before
    * @see #execute
    */
   @Override
   public void after(final FWDEngineExecutionContext context) throws Exception
   {
      for (final Method m : after)
      {
         // System.err.println("executing " + m + " after " + method + " on " + testInstance);
         ReflectionSupport.invokeMethod(m, testInstance);
      }
   }

   /**
    * Execute the <em>before</em> behavior of this node.
    *
    * <p>This method will be called once <em>before</em> {@linkplain #execute
    * execution} of this node.
    *
    * @param context
    *        the context to execute in
    * @return the new context to be used for children of this node; never
    * {@code null}
    * 
    * @see #execute
    * @see #after
    */
   @Override
   public FWDEngineExecutionContext before(FWDEngineExecutionContext context) throws Exception
   {
      for (final Method m : before)
      {
         // System.err.println("executing " + m + " + before " + method + " on " + testInstance);
         ReflectionSupport.invokeMethod(m, testInstance);
      }

      return context;
   }

   /**
    * Clean up the supplied {@code context} after execution.
    * 
    * Reset the test class instance.
    *
    * @param context
    *        the context to execute in
   
    * @see #prepare
    */
   @Override
   public void cleanUp(final FWDEngineExecutionContext context) throws Exception
   {
      testInstance = null;
   }

   /**
    * Execute the <em>behavior</em> of this node.
    *
    * <p>Containers typically do not implement this method since the
    * {@link HierarchicalTestEngine} handles execution of their children.
    *
    * <p>The supplied {@code dynamicTestExecutor} may be used to submit
    * additional dynamic tests for immediate execution.
    *
    * @param context
    *         the context to execute in
    * @param dynamicTestExecutor
    *        the executor to submit dynamic tests to
    *        
    * @return the new context to be used for children of this node and for the
    * <em>after</em> behavior of the parent of this node, if any
    * 
    * @see #before
    * @see #after
    */
   @Override
   public FWDEngineExecutionContext execute(final FWDEngineExecutionContext context,
            DynamicTestExecutor dynamicTestExecutor) throws Exception
   {
      // System.err.println("executing method " + method + " on " + testInstance);
      ReflectionSupport.invokeMethod(method, testInstance);

      return context;
   }

   /**
    * Determine the {@link org.junit.platform.engine.TestDescriptor.Type} of this descriptor.
    *
    * @return the descriptor type; never {@code null}.
    * @see #isContainer()
    * @see #isTest()
    */
   @Override
   public Type getType()
   {
      return Type.TEST;
   }

   /**
    * Prepare the supplied {@code context} prior to execution.
    * 
    * Create test class instance.
    *
    * @see #cleanUp
    */
   @Override
   public FWDEngineExecutionContext prepare(FWDEngineExecutionContext context) throws Exception
   {
      testInstance = ReflectionSupport.newInstance(clazz);

      return context;
   }

   /**
    * Get the method.
    * 
    * @return the method, never {@code null}
    */
   final Method getMethod()
   {
      return method;
   }
}
