package com.goldencode.p2j.testengine;

import java.util.*;

import org.junit.platform.commons.support.*;
import org.junit.platform.engine.*;
import org.junit.platform.engine.discovery.*;
import org.junit.platform.engine.support.hierarchical.*;
import org.junit.platform.suite.api.*;

/**
 * FWD test engine. 
 *
 */
public class FWDTestEngine
extends HierarchicalTestEngine<FWDEngineExecutionContext>
implements Node<FWDEngineExecutionContext>
{

   /**
    * Scan a class for test methods, declared in the class.
    * 
    * @param engineDescriptor
    *        the engine descriptor to register found test descriptors in
    * @param javaClass
    *        the Java class to scan
    */
   private final static void addOptionally(final TestDescriptor engineDescriptor,
            final Class<?> javaClass)
   {
      final FWDClassTestDescriptor instance = FWDClassTestDescriptor.testAndCreate(javaClass,
               engineDescriptor);
      if (instance != null)
      {
         engineDescriptor.addChild(instance);
      }
   }

   /**
    * Scan a class for all test methods, including whose defined in the class and referenced classes,
    * if the class is marked as test suite.
    * 
    * @param engineDescriptor
    *        the engine descriptor to register found test descriptors in
    * @param javaClass
    *        the Java class to scan
    */
   private static void scanClass(final TestDescriptor engineDescriptor, final Class<?> javaClass)
   {
      addOptionally(engineDescriptor, javaClass);
      // Add suites
      if (AnnotationSupport.isAnnotated(javaClass, Suite.class))
      {
         for (final Class<?> referencedClass : javaClass.getAnnotation(SelectClasses.class).value())
         {
            addOptionally(engineDescriptor, referencedClass);
         }

         // TODO: implement all other suite selectors if ever required...
      }
   }

   /**
    * Discover tests according to the supplied {@link EngineDiscoveryRequest}.
    *
    * <p>The supplied {@link UniqueId} must be used as the unique ID of the
    * returned root {@link TestDescriptor}. In addition, the {@code UniqueId}
    * must be used to create unique IDs for children of the root's descriptor
    * by calling {@link UniqueId#append}.
    *
    * @param discoveryRequest the discovery request; never {@code null}
    * @param uniqueId the unique ID to be used for this test engine's
    * {@code TestDescriptor}; never {@code null}
    * @return the root {@code TestDescriptor} of this engine, typically an
    * instance of {@code EngineDescriptor}
    * @see org.junit.platform.engine.support.descriptor.EngineDescriptor
    */
   @Override
   public TestDescriptor discover(final EngineDiscoveryRequest discoveryRequest, final UniqueId uniqueId)
   {
      final TestDescriptor engineDescriptor = new FWDEngineDescriptor(uniqueId);

      discoveryRequest.getSelectorsByType(ClassSelector.class).forEach(selector -> {
         scanClass(engineDescriptor, selector.getJavaClass());
      });

      discoveryRequest.getSelectorsByType(PackageSelector.class).forEach(selector -> {
         final List<Class<?>> classes = ReflectionSupport.findAllClassesInPackage(
                  selector.getPackageName(), (Class<?> c) -> true, name -> true);
         for(Class<?> javaClass : classes)
         {
            scanClass(engineDescriptor, javaClass);
         }
      });

      return engineDescriptor;
   }

   /**
    * Get the ID that uniquely identifies this test engine.
    *
    * <p>Each test engine must provide a unique ID. For example, JUnit Vintage
    * and JUnit Jupiter use {@code "junit-vintage"} and {@code "junit-jupiter"},
    * respectively. When in doubt, you may use the fully qualified name of your
    * custom {@code TestEngine} implementation class.
    */
   @Override
   public String getId()
   {
      return "junit-fwd";
   }

   /**
    * Create the initial execution context for executing the supplied
    * {@linkplain ExecutionRequest request}.
    *
    * @param request the request about to be executed
    * @return the initial context that will be passed to nodes in the hierarchy
    */
   @Override
   protected FWDEngineExecutionContext createExecutionContext(ExecutionRequest request)
   {
      return new FWDEngineExecutionContext();
   }

}
